#include "session.hh"
#include "window.hh"
#include "blackbox.hh"
#include "workspace.hh"
#include "icon.hh"

#include "icon.xpm"

#include <signal.h>
#include <string.h>
#include <stdlib.h>

XColor *colors_8bpp = 0;


SessionMenu::SessionMenu(BlackboxSession *s) : BlackboxMenu(s)
{ }
int SessionMenu::insert(char *l, int f, char *e)
{ return BlackboxMenu::insert(l, f, e); }
void SessionMenu::showMenu(void)
{ BlackboxMenu::showMenu(); }
void SessionMenu::updateMenu(void)
{ BlackboxMenu::updateMenu(); }
void SessionMenu::itemPressed(int, int) { }
void SessionMenu::titlePressed(int) { }
void SessionMenu::titleReleased(int button) { if (button == 3) hideMenu(); }


void SessionMenu::itemReleased(int button, int index) {
  if (button == 1) {
    BlackboxMenuItem *item = at(index);
    if (item->Exec()) {
      char *command = new char[strlen(item->Exec()) + 8];
      sprintf(command, "exec %s &", item->Exec());
      system(command);
      delete [] command;
    } else if (item->Function()) {
      switch (item->Function()) {
      case BlackboxSession::B_Restart:
	Session()->Restart();
	break;

      case BlackboxSession::B_Exit:
	Session()->Exit();
	break;
      }
    }
  }
}


/*

  Resources allocated for each session:
  linked list - ilist
  Debugger - debug
  XFontStruct = resource font title, menu
  XpmAttributes - icon default attrib
  GC - opGC
  WorkspaceManager - ws_manager
  SessionMenu - rootmenu

*/

BlackboxSession::BlackboxSession(char *display_name) {
  signal(SIGSEGV, (void (*)(int)) sig11handler);

  b1Pressed = False;
  b2Pressed = False;
  b3Pressed = False;
  startup = True;

  ilist = new llist<BlackboxIcon>;
  debug = new Debugger();
#ifdef DEBUG
  debug->enable();
#endif
  debug->enter("BlackboxSession::Blackbox\n");

  debug->msg("opening X server connection\n");
  if ((display = XOpenDisplay(display_name)) == NULL) {
    fprintf(stderr, "connection to X server failed\n");
    exit(2);
  }
  else
    debug->msg("opened to '%s'... beginning management\n",
	       XDisplayName(display_name));
  
  screen = DefaultScreen(display);
  root = RootWindow(display, screen);
  v = DefaultVisual(display, screen);
  depth = DefaultDepth(display, screen);

#ifdef SHAPE
  {
    debug->msg("querying X server shape extensions\n");
    int i, ii;
    if (! (shape_extensions = XShapeQueryExtension(display, &i, &ii)))
      debug->msg("server doesn't support shape extensions\n");
    else
      debug->msg("server supports shape extensions\n"
		 "   (event_basep %d, error_basep %d)\n", i, ii);
  }
#else
  shape_extensions = False;
#endif

  LoadDefaults();

  debug->msg("loading default font\n");
  const char *defaultFont = "-*-charter-medium-r-*-*-*-120-*-*-*-*-*-*";
  if ((resource.font.title = XLoadQueryFont(display, defaultFont)) == NULL) {
    fprintf(stderr,
	    "blackbox: couldn't load default font.  please check to\n"
	    "make sure the necessary font is installed '%s'\n", defaultFont);
    exit(2);
  }
  resource.font.menu = resource.font.title;

  InitScreen();
}



/*

  Resources deallocated for each session:
  linked list - ilist... the icon data is deallocated by each workspaces
      Dissociate() call

  Debugger - debug
  XFontStruct = resource font title, menu
  XpmAttributes - icon default attrib
  GC - opGC
  WorkspaceManager - ws_manager
  SessionMenu - rootmenu

*/

BlackboxSession::~BlackboxSession() {
  XSelectInput(display, root, NoEventMask);

  XFreeFont(display, resource.font.title);
  // XFreeFont(display, resource.font.menu);
  XFreeGC(display, opGC);
  
  delete rootmenu;
  delete ws_manager;
  delete icon.default_attrib;
  delete ilist;
  debug->msg("closing X connection\n");
  XSync(display, 0);
  XCloseDisplay(display);
  delete debug;
}


int BlackboxSession::anotherWMRunning(Display *, XErrorEvent *) {
  fprintf(stderr,
	  "blackbox: a fatal error has occurred while querying the X server\n"
	  "          make sure there is not another window manager running\n");
  exit(3);
  return(-1);
}


void BlackboxSession::InitScreen(void) {
  context.app = XUniqueContext();
  context.window = XUniqueContext();
  context.icon = XUniqueContext();
  context.menu = XUniqueContext();

  _XA_WM_COLORMAP_WINDOWS = XInternAtom(display, "WM_COLORMAP_WINDOWS", False);
  _XA_WM_PROTOCOLS = XInternAtom(display, "WM_PROTOCOLS", False);
  _XA_WM_STATE = XInternAtom(display, "WM_STATE", False);
  _XA_WM_DELETE_WINDOW = XInternAtom(display, "WM_DELETE_WINDOW", False);
  _XA_WM_TAKE_FOCUS = XInternAtom(display, "WM_TAKE_FOCUS", False);

  event_mask = LeaveWindowMask | EnterWindowMask | PropertyChangeMask |
    SubstructureNotifyMask | PointerMotionMask | SubstructureRedirectMask |
    ButtonPressMask | ButtonReleaseMask | KeyPressMask | KeyReleaseMask;
  
  XSetErrorHandler((XErrorHandler) anotherWMRunning);
  XSelectInput(display, root, event_mask);
  XSync(display, 0);
  XSetErrorHandler((XErrorHandler) NULL);

  cursor.session = XCreateFontCursor(display, XC_left_ptr);
  cursor.move = XCreateFontCursor(display, XC_fleur);
  XDefineCursor(display, root, cursor.session);

  xres = WidthOfScreen(ScreenOfDisplay(display, screen));
  yres = HeightOfScreen(ScreenOfDisplay(display, screen));

  debug->msg("creating default icon pixmap\n");
  icon.default_attrib = new XpmAttributes;
  bzero((void *) icon.default_attrib, sizeof(XpmAttributes));

  int ret;
  if ((ret = XpmCreatePixmapFromData(display, root, blackbox_icon,
				     &icon.default_icon, &icon.default_mask,
				     icon.default_attrib)) != XpmSuccess) {
    debug->msg("error creating default icon %d\n", ret);
    icon.default_icon = None;
    icon.default_mask = None;
  }
  
  if (depth == 8) {
    colors_8bpp = new XColor[216];
    int i = 0;
    for (int r = 0; r < 6; ++r)
      for (int g = 0; g < 6; ++g)
	for (int b = 0; b < 6; ++b) {
	  colors_8bpp[i].red = (r * 0xffff) / 5;
	  colors_8bpp[i].green = (g * 0xffff) / 5;
	  colors_8bpp[i].blue = (b * 0xffff) / 5;
	  colors_8bpp[i++].flags = DoRed|DoGreen|DoBlue;
	}
    
    XGrabServer(display);
    
    Colormap colormap = DefaultColormap(display, DefaultScreen(display));
    for (i = 0; i < 216; ++i)
      if (! XAllocColor(display, colormap, &colors_8bpp[i])) {
	fprintf(stderr, "couldn't alloc color %i %i %i\n", colors_8bpp[i].red,
		colors_8bpp[i].green, colors_8bpp[i].blue);		
	colors_8bpp[i].flags = 0;
      } else
	colors_8bpp[i].flags = DoRed|DoGreen|DoBlue;
    
    int close_colorsi = ((1 << depth) > 256) ? 256 : 1 << depth;
    XColor close_colorsa[close_colorsi];
    
    for (i = 0; i < close_colorsi; ++i) close_colorsa[i].pixel = i;
    XQueryColors(display, colormap, close_colorsa, close_colorsi);
    
    for (i = 0; i < 216; ++i) {
      if (colors_8bpp[i].flags == 0) {
	int ii, r = 2, rr, gg, bb;
	unsigned long c = 0xffffffff, d, l = 0l;
	
	while (r--) {
	  for (ii = 0; ii < close_colorsi; ++ii) {
	    rr = (colors_8bpp[i].red - close_colorsa[i].red) >> 8;
	    gg = (colors_8bpp[i].green - close_colorsa[i].green) >> 8;
	    bb = (colors_8bpp[i].blue - close_colorsa[i].blue) >> 8;
	    d = rr*rr + gg*gg + bb*bb;
	    if (d < c) {
	      c = d;
	      l = ii;
	    }
	  }
	  
	  colors_8bpp[i].red = close_colorsa[l].red;
	  colors_8bpp[i].green = close_colorsa[l].green;
	  colors_8bpp[i].blue = close_colorsa[l].blue;
	  if (XAllocColor(display, colormap, &colors_8bpp[i])) {
	    colors_8bpp[i].flags = DoRed|DoGreen|DoBlue;
	    break;
	  }
	}
      }
    }
      
    XUngrabServer(display);
  } else
    colors_8bpp = 0;
  
  debug->msg("setting icon size hint\n");
  XIconSize iconsize;
  iconsize.min_width = 1;
  iconsize.min_height = 1;
  iconsize.max_width = 96;
  iconsize.max_height = 96;
  iconsize.width_inc = 1;
  iconsize.height_inc = 1;
  XSetIconSizes(display, root, &iconsize, 1);

  XGCValues gcv;
  gcv.foreground = getColor("grey");
  gcv.function = GXxor;
  gcv.line_width = 2;
  gcv.subwindow_mode = IncludeInferiors;
  gcv.font = resource.font.title->fid;
  opGC = XCreateGC(display, root, GCForeground|GCFunction|GCSubwindowMode|
		   GCFont, &gcv);

  InitMenu();
  ws_manager = new WorkspaceManager(this);

  int i;
  unsigned int nchild;
  Window r, p, *children;
  XGrabServer(display);
  XQueryTree(display, root, &r, &p, &children, &nchild);

  debug->msg("reparenting %u existing window(s)\n", nchild);
  debug->msg("tree: %lx - %lx - %lx\n", root, r, p);

  for (i = 0; i < (int) nchild; ++i) {
    if (children[i] == None) continue;

    XWindowAttributes attrib;
    if (XGetWindowAttributes(display, children[i], &attrib)) {
      if ((! attrib.override_redirect) && (attrib.map_state != IsUnmapped)) {
	XSync(display, 0);
	BlackboxWindow *nWin = new BlackboxWindow(this, root, children[i]);
		
	Atom atom;
	int foo;
	unsigned long ulfoo, nitems;
	unsigned char *state;
	XGetWindowProperty(display, nWin->clientWindow(), _XA_WM_STATE,
			   0, 3, False, _XA_WM_STATE, &atom, &foo, &nitems,
			   &ulfoo, &state);

	if (state != NULL) {
	  switch (*((unsigned long *) state)) {
	  case WithdrawnState:
	    nWin->deiconifyWindow();
            nWin->setFocusFlag(False);
	    break;
	    
	  case IconicState:
	    nWin->iconifyWindow();
	    break;
	    
	  case NormalState:
	  default:
	    nWin->deiconifyWindow();
	    nWin->setFocusFlag(False);
	    break;
	  }
	} else {
	  nWin->deiconifyWindow();
	  nWin->setFocusFlag(False);
	} 
      } else
	debug->msg("do not reparent override redirect windows\n");
      
    } else
      debug->msg("couldn't get window attributes\n");
  }

  while (XPending(display)) {
    XEvent foo;
    XNextEvent(display, &foo);
  }

  XUngrabServer(display);
}


void BlackboxSession::EventLoop(void) {
  shutdown = False;
  startup = False;
#ifdef DEBUG
  XSynchronize(display, True);
#endif

  for (; (! shutdown);) {
    XEvent e;
    XNextEvent(display, &e);

    ProcessEvent(&e);
  }

  Dissociate();
}


void BlackboxSession::InitMenu(void) {
  rootmenu = new SessionMenu(this);

  char *line = new char[121], *label = new char[41], *command = new char[81];
  FILE *menu_file = fopen(resource.menuFile, "r");
  if (menu_file != NULL) {
    if (! feof(menu_file)) {
      fgets(line, 120, menu_file);
      int len = strlen(line);
      strncpy(command, line, 5);
      command[5] = '\0';
      strncpy(label, line + 5, len - 5);

      if (! strncasecmp(command, "begin", 5)) {
        int i, ri, llen = strlen(label);
        for (i = 0; i < llen; ++i)
          if (label[i] == '(') { ++i; break; }

        for (ri = llen; ri > 0; --ri)
          if (label[ri] == ')') break;
	
        if (i < ri && ri > 0) {
          char *newlabel = new char[ri - i + 1];
          strncpy(newlabel, label + i, ri - i);
          *(newlabel + (ri - i)) = '\0';
          rootmenu->setMenuLabel(newlabel);
        } 

        while (! feof(menu_file)) {
          fgets(line, 120, menu_file);
          if (! strncasecmp(line, "end", 3)) break;
          else {
            len = strlen(line);
            for (i = 0; i < len; ++i)
              if (line[i] == '\"') { ++i; break; }
            for (ri = len; ri > 0; --ri)
              if (line[ri] == '\"') break;

            char *l;
            if (i < ri && ri > 0) {
              l = new char[ri - i + 1];
              strncpy(l, line + i, ri - i);
              *(l + (ri - i)) = '\0';
            } else
              l = "(nil)";

            for (i = 0; i < len; ++i)
              if (line[i] == '(') { ++i; break; }
            for (ri = len; ri > 0; --ri)
              if (line[ri] == ')') break;

            char *c;
            if (i < ri && ri > 0) {
              c = new char[ri - i + 1];
              strncpy(c, line + i, ri - i);
              *(c + (ri - i)) = '\0';
            } else
              c = 0;

            if (c) {
              if (! strncasecmp(c, "Restart", 7)) {
                rootmenu->insert(l, B_Restart);
                delete [] c;
              } else if (! strncasecmp(c, "Exit", 4)) {
                rootmenu->insert(l, B_Exit);
                delete [] c;
              } else if (! strncasecmp(c, "Shutdown", 8)) {
                rootmenu->insert(l, B_Shutdown);
                delete [] c;
              } else
                rootmenu->insert(l, B_Execute, c);
            } else
              fprintf(stderr, "error in menu file... label must have command\n"
                      "  ex:  \"label\" (command)\n");
          }
        }
      } else {
        
      }

    } else {
      perror(resource.menuFile);
      // error in menu file
    }

    fclose(menu_file);
  } else {
    // no menu file... fall back on default
    perror(resource.menuFile);
    rootmenu->insert("xterm", B_Execute, "xterm");
    rootmenu->insert("Restart", B_Restart);
    rootmenu->insert("Exit", B_Exit);
  }

  delete [] command;
  delete [] line;
  delete [] label;
  rootmenu->updateMenu();
}


Pixel BlackboxSession::getColor(const char *colorname) {
  XColor color;
  XWindowAttributes attributes;
  
  XGetWindowAttributes(display, root, &attributes);
  color.pixel = 0;
  if (!XParseColor(display, attributes.colormap, colorname, &color)) {
    fprintf(stderr, "blackbox: color parse error: %s", colorname);
  } else if (!XAllocColor(display, attributes.colormap, &color)) {
    fprintf(stderr, "blackbox: color alloc error: %s", colorname);
  }

  return color.pixel;
}


void BlackboxSession::ProcessEvent(XEvent *e) {
  switch (e->type) {
  case ButtonPress: {
    BlackboxWindow *bWin = NULL;
    BlackboxIcon *bIcon = NULL;
    BlackboxMenu *bMenu = NULL;

    switch (e->xbutton.button) {
    case 1: b1Pressed = True; break;
    case 2: b2Pressed = True; break;
    case 3: b3Pressed = True; break;
    }

    debug->msg("button press %ld\n", e->xbutton.time);

    if ((bWin = getWindow(display, e->xbutton.window)) != NULL) {
      bWin->buttonPressEvent(&e->xbutton);
    } else if ((bIcon = getIcon(display, e->xbutton.window)) != NULL) {
      bIcon->buttonPressEvent(&e->xbutton);
    } else if ((bMenu = getMenu(display, e->xbutton.window)) != NULL) {
      bMenu->buttonPressEvent(&e->xbutton);
    } else if (e->xbutton.window == root && e->xbutton.button == 1) {
      ws_manager->showMenu();
    } else if (e->xbutton.window == root && e->xbutton.button == 2) {
      ws_manager->currentWorkspace()->showMenu();
    } else if (e->xbutton.window == root && e->xbutton.button == 3) {
      rootmenu->showMenu();
    }

    break;
  }
  
  case ButtonRelease: {
    BlackboxWindow *bWin = NULL;
    BlackboxIcon *bIcon = NULL;
    BlackboxMenu *bMenu = NULL;

    switch (e->xbutton.button) {
    case 1: b1Pressed = False; break;
    case 2: b2Pressed = False; break;
    case 3: b3Pressed = False; break;
    }

    if ((bWin = getWindow(display, e->xbutton.window)) != NULL)
      bWin->buttonReleaseEvent(&e->xbutton);
    else if ((bIcon = getIcon(display, e->xbutton.window)) != NULL)
      bIcon->buttonReleaseEvent(&e->xbutton);
    else if ((bMenu = getMenu(display, e->xbutton.window)) != NULL)
      bMenu->buttonReleaseEvent(&e->xbutton);
    
    break;
  }
  
  case ConfigureRequest: {
    BlackboxWindow *cWin = getWindow(display, e->xconfigurerequest.window);
    if (cWin != NULL)
      cWin->configureRequestEvent(&e->xconfigurerequest);
    else {
      /* configure a window we haven't mapped yet */
      debug->msg("configuring unmapped window\n");
      XWindowChanges xwc;
      
      xwc.x = e->xconfigurerequest.x;
      xwc.y = e->xconfigurerequest.y;	
      xwc.width = e->xconfigurerequest.width;
      xwc.height = e->xconfigurerequest.height;
      xwc.border_width = 0;
      xwc.sibling = e->xconfigurerequest.above;
      xwc.stack_mode = e->xconfigurerequest.detail;
      
      XConfigureWindow(display, e->xconfigurerequest.window,
		       e->xconfigurerequest.value_mask, &xwc);
    }
    
    break;
  }
    
  case MapRequest: {
    BlackboxWindow *rWin = getWindow(display, e->xmaprequest.window);
    if (rWin == NULL)
	rWin = new BlackboxWindow(this, root, e->xmaprequest.window);
    
    rWin->mapRequestEvent(&e->xmaprequest);
    
      break;
  }
  
  case MapNotify: {
    debug->msg("map notify\n");
    BlackboxWindow *mWin = getWindow(display, e->xmap.window);
    if (mWin != NULL)
      mWin->mapNotifyEvent(&e->xmap);
    
    break;
  }
  
  case UnmapNotify: {
    debug->msg("unmap notify\n");
    BlackboxWindow *uWin = getWindow(display, e->xunmap.window);
    if (uWin != NULL)
      uWin->unmapNotifyEvent(&e->xunmap);
    
    break;
  }
  
  case DestroyNotify: {
    BlackboxWindow *dWin = getWindow(display, e->xdestroywindow.window);
    if (dWin != NULL)
      dWin->destroyNotifyEvent(&e->xdestroywindow);
    
    break;
  }
  
  case MotionNotify: {
    BlackboxWindow *mWin = NULL;
    BlackboxIcon *mIcon = NULL;
    BlackboxMenu *mMenu = NULL;
    if ((mWin = getWindow(display, e->xmotion.window)) != NULL)
      mWin->motionNotifyEvent(&e->xmotion);
    else if ((mIcon = getIcon(display, e->xmotion.window)) != NULL)
      mIcon->motionNotifyEvent(&e->xmotion);
    else if ((mMenu = getMenu(display, e->xmotion.window)) != NULL)
      mMenu->motionNotifyEvent(&e->xmotion);
    
    break;
  }
  
  case PropertyNotify: {
    char *atomname = XGetAtomName(display, e->xproperty.atom);
    debug->msg("property notify %s\n", atomname);
    XFree(atomname);

    if (e->xproperty.state != PropertyDelete) {
      BlackboxWindow *pWin = getWindow(display, e->xproperty.window);
      if (pWin != NULL)
	pWin->propertyNotifyEvent(e->xproperty.atom);
    }

    break;
  }
  
  case EnterNotify: {
    BlackboxWindow *fWin = NULL;
    BlackboxIcon *fIcon = NULL;
    
    if ((fWin = getWindow(display, e->xcrossing.window)) != NULL) {
      XGrabServer(display);
      
      XSync(display, 0);
      XEvent foo;
      if (XCheckTypedWindowEvent(display, fWin->clientWindow(),
				 UnmapNotify, &foo)) {
	debug->msg("unmap notify event enqueued for window %lx\n"
		   "processing unmap event\n",
		   fWin->clientWindow());
	ProcessEvent(&foo);
      } else if ((! fWin->isFocused()) && fWin->isVisible()) {
	fWin->setInputFocus();
      }
      
      XUngrabServer(display);
    } else if ((fIcon = getIcon(display, e->xcrossing.window)) != NULL)
      fIcon->enterNotifyEvent(&e->xcrossing);
    
    break;
  }
  
  case LeaveNotify: {
    BlackboxIcon *lIcon = NULL;
    if ((lIcon = getIcon(display, e->xcrossing.window)) != NULL)
      lIcon->leaveNotifyEvent(&e->xcrossing);
    
    break;
  }
  
  case Expose: {
    BlackboxWindow *eWin = NULL;
    BlackboxMenu *eMenu = NULL;
    if ((eWin = getWindow(display, e->xexpose.window)) != NULL)
      eWin->exposeEvent(&e->xexpose);
    else if ((eMenu = getMenu(display, e->xexpose.window)) != NULL)
      eMenu->exposeEvent(&e->xexpose);
    
    break;
  } 
  
  case FocusIn: {
    BlackboxWindow *iWin = getWindow(display, e->xfocus.window);
    if (iWin != NULL)
      iWin->setFocusFlag(True);
    
    break;
  }
  
  case FocusOut: {
    BlackboxWindow *oWin = getWindow(display, e->xfocus.window);
    if (oWin != NULL)
      oWin->setFocusFlag(False);
    
    break;
  }
  
  // the events we do not handle
  case CreateNotify:
  case ReparentNotify:
  case KeyPress:
  case KeyRelease:
  case ConfigureNotify:
    break;
    
  default:
    debug->msg("event %d: %lx\n", e->type, e->xany.window);
    break;
  }
}


BlackboxWindow *BlackboxSession::getWindow(Display *d, Window window) {
  BlackboxWindow *win = NULL;
  XEvent foo;

  if (XCheckTypedWindowEvent(display, window, DestroyNotify, &foo))
    ProcessEvent(&foo);
  else if (XFindContext(d, window, context.window, (XPointer *) &win)
	   != XCNOENT)
    return win;
  
  return NULL;
}


BlackboxIcon *BlackboxSession::getIcon(Display *d, Window window) {
  BlackboxIcon *icon = NULL;
  XEvent foo;
  
  if (XCheckTypedWindowEvent(display, window, DestroyNotify, &foo))
    ProcessEvent(&foo);
  else if (XFindContext(d, window, context.icon, (XPointer *) &icon)
	   != XCNOENT)
    return icon;
  
  return NULL;
}


BlackboxMenu *BlackboxSession::getMenu(Display *d, Window window) {
  BlackboxMenu *menu = NULL;
  XEvent foo;
  
  if (XCheckTypedWindowEvent(display, window, DestroyNotify, &foo))
    ProcessEvent(&foo);
  else if (XFindContext(d, window, context.menu, (XPointer *) &menu)
	   != XCNOENT)
    return menu;
  
  return NULL;
}


void BlackboxSession::arrangeIcons(void) {
  int i;
  BlackboxIcon *icon = NULL;

  XGrabServer(display);
  for (i = 0; i < ilist->count(); ++i) {
    debug->msg("positioning icon %d\n", i);
    icon = ilist->at(i);
    XMoveWindow(display, icon->iconWindow(), xres - icon->Width() - 2,
		      (icon->Height() + 2) * (ilist->count() - (i + 1)));
  }

  XUngrabServer(display);
}


void BlackboxSession::Dissociate(void) {
  debug->msg("shutting down blackbox...\n"
	     "discontinuing management of %d workspace(s)\n",
	     ws_manager->count());
  ws_manager->DissociateAll();
}


void BlackboxSession::Restart(void) {
  Dissociate();
  XSetInputFocus(display, PointerRoot, RevertToParent, CurrentTime);
  blackbox->Restart();
}


void BlackboxSession::Exit(void) {
  XSetInputFocus(display, PointerRoot, RevertToParent, CurrentTime);
  shutdown = True;  
}


void BlackboxSession::addWindow(BlackboxWindow *w)
{ ws_manager->currentWorkspace()->addWindow(w); }


void BlackboxSession::removeWindow(BlackboxWindow *w)
{ ws_manager->workspace(w->workspace())->removeWindow(w); }


int BlackboxSession::errorHandler(Display *d, XErrorEvent *e) {
  fprintf(stderr, "error in window manager\n");

  char err[80];
  XGetErrorText(d, e->error_code, err, 79);
  fprintf(stderr, "error code %d - %s\n", e->error_code, err);
  fprintf(stderr, "major/minor opcode - %d/%d XID %lx\n", e->request_code,
    e->minor_code, e->resourceid);

  Dissociate();
  exit(1);
}

void BlackboxSession::sig11handler(int i) {
  static int re_enter = 0;

  fprintf(stderr, "%d sig 11 caught...", i);
  if (! re_enter) {
    re_enter = 1;
    fprintf(stderr, "\ndissociating windows...");
    Dissociate();
  }

  fprintf(stderr, "exiting...\n");
  abort();
}


void BlackboxSession::reassociateWindow(BlackboxWindow *w) {
  if (w->workspace() != ws_manager->currentWorkspaceID()) {
    ws_manager->workspace(w->workspace())->removeWindow(w);
    ws_manager->currentWorkspace()->addWindow(w);
  }
}


void BlackboxSession::LoadDefaults(void) {
  //
  // blackbox defaults...
  //

  debug->msg("assigning defaults\n");
  resource.color.frame.r = 0x00;
  resource.color.frame.g = 0x00;
  resource.color.frame.b = 0x00;
  resource.color.focus.r = 0xa0;
  resource.color.focus.g = 0xa0;
  resource.color.focus.b = 0xa0;
  resource.color.unfocus.r = 0x40;
  resource.color.unfocus.g = 0x40;
  resource.color.unfocus.b = 0x40;
  resource.color.button.r = 0xa0;
  resource.color.button.g = 0xa0;
  resource.color.button.b = 0xa0;
  resource.color.icon.r = 0x00;
  resource.color.icon.g = 0x00;
  resource.color.icon.b = 0x00;
  resource.color.menu.r = 0xa0;
  resource.color.menu.g = 0xa0;
  resource.color.menu.b = 0xa0;
  resource.color.imenu.r = 0x40;
  resource.color.imenu.g = 0x40;
  resource.color.imenu.b = 0x40;
  resource.color.imenu_to.r = 0x00;
  resource.color.imenu_to.g = 0x00;
  resource.color.imenu_to.b = 0x00;
#define BLACKBOXMENUAD XLIBDIR##"/Blackbox/MenuAD"
  resource.menuFile = BLACKBOXMENUAD;

  debug->msg("initializing Xrm\n");
  XrmInitialize();

  debug->msg("retrieving databases\n");
#define BLACKBOXAD XAPPLOADDIR##"/BlackboxAD"
  XrmDatabase default_database = XrmGetDatabase(display),
    blackbox_database = XrmGetFileDatabase(BLACKBOXAD),
    resource_database = NULL;

  if (XResourceManagerString(display) != NULL) {
    debug->msg("loading resource manager database\n");
    resource_database = XrmGetStringDatabase(XResourceManagerString(display));
  }
    
  if (resource_database != NULL) {
    debug->msg("combining resource manager database\n");
    XrmCombineDatabase(resource_database, &default_database, True);
  }

  if (blackbox_database == NULL && default_database == NULL)
    return;

  debug->msg("combining databases %p %p\n", default_database,
	     blackbox_database);
  XrmCombineDatabase(default_database, &blackbox_database, True);
  
  //
  // default_database is destroyed by the above call...
  //
  // let's continue parsing our default resources
  //

  XrmValue value;
  char *value_type;

  if (XrmGetResource(blackbox_database, "blackbox.frameColor",
		     "Blackbox.FrameColor", &value_type, &value)) {
    int r, g, b;
    if ((sscanf(value.addr, "\"rgb:%x/%x/%x\"", &r, &g, &b) == 3) ||
	(sscanf(value.addr, "rgb:%x/%x/%x", &r, &g, &b) == 3)) {
      resource.color.frame.r = r;
      resource.color.frame.g = g;
      resource.color.frame.b = b;
    }
  }

  if (XrmGetResource(blackbox_database,
		     "blackbox.associatedWindow.titleTexture",
		     "Blackbox.AssociatedWindow.TitleTexture",
		     &value_type, &value)) {
    if ((! strcasecmp(value.addr, "solid")) ||
	(! strcasecmp(value.addr, "solidraised")))
      resource.texture.title = B_TextureRSolid;
    else if ((!strcasecmp(value.addr, "solidsunken")))
      resource.texture.title = B_TextureSSolid;
    // else if ((!strcasecmp(value.addr, "dgradient")) ||
    //          (! strcasecmp(value.addr, "dgradientraised")))
    //   resource.texture.title = BTextureRDGradient;
  } else
    resource.texture.title = B_TextureRSolid;
  
  if (XrmGetResource(blackbox_database,
		     "blackbox.associatedWindow.focusColor",
		     "Blackbox.AssociatedWindow.FocusColor", &value_type,
		     &value)) {
    int r, g, b;
    if ((sscanf(value.addr, "\"rgb:%x/%x/%x\"", &r, &g, &b) == 3) ||
	(sscanf(value.addr, "rgb:%x/%x/%x", &r, &g, &b) == 3)) {
      resource.color.focus.r = r;
      resource.color.focus.g = g;
      resource.color.focus.b = b;
    }
  }

  if (XrmGetResource(blackbox_database,
		     "blackbox.associatedWindow.unfocusColor",
		     "Blackbox.AssociatedWindow.UnfocusColor", &value_type,
		     &value)) {
    int r, g, b;
    if ((sscanf(value.addr, "\"rgb:%x/%x/%x\"", &r, &g, &b) == 3) ||
	(sscanf(value.addr, "rgb:%x/%x/%x", &r, &g, &b) == 3)) {
      resource.color.unfocus.r = r;
      resource.color.unfocus.g = g;
      resource.color.unfocus.b = b;
    }
  }

  if (XrmGetResource(blackbox_database,
		     "blackbox.associatedWindow.buttonColor",
		     "Blackbox.AssociatedWindow.ButtonColor", &value_type,
		     &value)) {
    int r, g, b;
    if ((sscanf(value.addr, "\"rgb:%x/%x/%x\"", &r, &g, &b) == 3) ||
	(sscanf(value.addr, "rgb:%x/%x/%x", &r, &g, &b) == 3)) {
      resource.color.button.r = r;
      resource.color.button.g = g;
      resource.color.button.b = b;
    }
  }

  if (XrmGetResource(blackbox_database,
		     "blackbox.associatedWindow.iconColor",
		     "Blackbox.AssociatedWindow.IconColor", &value_type,
		     &value)) {
    int r, g, b;
    if ((sscanf(value.addr, "\"rgb:%x/%x/%x\"", &r, &g, &b) == 3) ||
	(sscanf(value.addr, "rgb:%x/%x/%x", &r, &g, &b) == 3)) {
      resource.color.icon.r = r;
      resource.color.icon.g = g;
      resource.color.icon.b = b;
    }
  }

  if (XrmGetResource(blackbox_database,
		     "blackbox.associatedWindow.menuColor",
		     "Blackbox.AssociatedWindow.MenuColor", &value_type,
		     &value)) {
    int r, g, b;
    if ((sscanf(value.addr, "\"rgb:%x/%x/%x\"", &r, &g, &b) == 3) ||
	(sscanf(value.addr, "rgb:%x/%x/%x", &r, &g, &b) == 3)) {
      resource.color.menu.r = r;
      resource.color.menu.g = g;
      resource.color.menu.b = b;
    }
  }

  if (XrmGetResource(blackbox_database,
		     "blackbox.associatedWindow.menuItemColor",
		     "Blackbox.AssociatedWindow.MenuItemColor", &value_type,
		     &value)) {
    int r, g, b;
    if ((sscanf(value.addr, "\"rgb:%x/%x/%x\"", &r, &g, &b) == 3) ||
	(sscanf(value.addr, "rgb:%x/%x/%x", &r, &g, &b) == 3)) {
      resource.color.imenu.r = r;
      resource.color.imenu.g = g;
      resource.color.imenu.b = b;
    }
  }

  if (XrmGetResource(blackbox_database,
		     "blackbox.associatedWindow.menuItemToColor",
		     "Blackbox.AssociatedWindow.MenuItemToColor", &value_type,
		     &value)) {
    int r, g, b;
    if ((sscanf(value.addr, "\"rgb:%x/%x/%x\"", &r, &g, &b) == 3) ||
	(sscanf(value.addr, "rgb:%x/%x/%x", &r, &g, &b) == 3)) {
      resource.color.imenu_to.r = r;
      resource.color.imenu_to.g = g;
      resource.color.imenu_to.b = b;
    }
  }

  if (XrmGetResource(blackbox_database,
		     "blackbox.sessionMenu.menuFile",
		     "Blackbox.SessionMenu.MenuFile", &value_type,
		     &value))
    resource.menuFile = value.addr;
  
  debug->msg("destroying database\n");
  XrmDestroyDatabase(blackbox_database);
}


unsigned long BlackboxSession::framePixel(void) {
  unsigned long pixel = 0l;

  switch (depth) {
  case 32:
  case 24:
    pixel = ((resource.color.frame.r << 16)|
	     (resource.color.frame.g << 8)|
	     (resource.color.frame.b));
    break;

  case 16:
    pixel = (((resource.color.frame.r << 8) & v->red_mask)|
	     ((resource.color.frame.g << 3) & v->green_mask)|
	     ((resource.color.frame.b >> 3) & v->blue_mask));
    break;

  default:
    pixel = 0l;
  }

  return pixel;
}


unsigned long BlackboxSession::iconPixel(void) {
  unsigned long pixel = 0l;

  switch (depth) {
  case 32:
  case 24:
    pixel = ((resource.color.icon.r << 16)|
	     (resource.color.icon.g << 8)|
	     (resource.color.icon.b));
    break;

  case 16:
    pixel = (((resource.color.icon.r << 8) & v->red_mask)|
	     ((resource.color.icon.g << 3) & v->green_mask)|
	     ((resource.color.icon.b >> 3) & v->blue_mask));
    break;

  default:
    pixel = 0l;
  }

  return pixel;
}


void BlackboxSession::updateWorkspace(int w) {
  ws_manager->workspace(w)->updateMenu();
}
